module RSArrowInstances(module RSArrowInstances, RS(..)) where
import Char(isSpace)
import Prec
import MonadTools((@@),liftM0,MonadPlus(..))
import Tools(bool2Either)
import RS(RS(..))
import ArrowClasses
import ArrowTools(StateArrT(..), MonadArrT(..), Kleisli(..),
                  ReaderArrT, updateStateArrT, parenthesize,
                  deparenthesize)
import MonadFunctorInstances()
import StateFunctorInstances()
import KleisliArrowInstances()

-- Trivial instances:

instance Arrow RS where
  arr f         = RS (arr f)
  RS f >>> RS g = RS (f >>> g)
  RS f <<< RS g = RS (f <<< g)
  first (RS f)  = RS (first f)
  second (RS f) = RS (second f)

instance ArrowChoice RS where
  RS f +++ RS g = RS (f +++ g)  
  RS f ||| RS g = RS (f ||| g)  

instance ArrowZero RS where
  zeroA = RS zeroA

instance ArrowPlus RS where
  RS f <|> RS g = RS (f <|> g)
  RS f <+> RS g = RS (f <+> g)

instance ReaderArrow Prec RS where
  peek          = RS peek
  poke f (RS g) = RS (poke f g)

instance StateArrow String RS where
  update f = RS (update f)

-- ----------------------------------------------------------------
-- Instances for ArrowSymbol, ArrowPrec and ArrowReadShow

instance ArrowSymbol RS where
  readSymbol = readSymbolRS
  showSymbol = showSymbolRS

instance (ArrowSymbol q, ReaderArrowChoice Prec q,
          ArrowPlus q) => ArrowPrec q where
  mayReadParen  = mayReadParenRS
  mayShowParen  = mayShowParenRS

  setPrec pri   = poke (const pri)

instance (ArrowChoice q, ArrowPlus q, ArrowSymbol q, ArrowPrec q) => 
  ArrowReadShow q 
 
-- ----------------------------------------------------------------
-- Implementations of the instances:

-- ArrowSymbol:
readSymbolRS :: String -> RS a a
readSymbolRS = readSymbolRSminimal

showSymbolRS :: StateArrow [c] q => [c] -> q a a
showSymbolRS sym = update (sym++)

readSymbolRSminimal :: String -> RS a a
readSymbolRSminimal sym = 
  RS (lift (updateStateArrT (Kleisli (readSymbolMinimal sym))))

readSymbolMinimal :: MonadPlus a => String -> String -> a String
readSymbolMinimal sym@(c:cs) =
  \s-> if null cs && c `elem` " ()" 
       then readToken c s
       else let (start,rest) = span (`notElem` " ()") s
            in if start == sym then return rest else mzero

-- ArrowPrec:

mayReadParenRS, mayShowParenRS :: 
  (ReaderArrowChoice Prec a, ArrowSymbol a) => 
  Prec -> a b c -> a b c
mayReadParenRS p f = ifHighPrec p (deparenthesize f) f
mayShowParenRS p f = ifHighPrec p   (parenthesize f) f

-- Utilities for ArrowPrec

ifHighPrec :: ReaderArrowChoice Prec a => 
  Prec -> a b c -> a b c -> a b c
ifHighPrec p = ifpeek (>p)

-- ----------------------------------------------------------------
-- Variants:

-- Simplest variants when the input is already tokenized
showSymbolConcrete :: 
  (Arrow q)           => s -> ReaderArrT r (StateArrT [s] q) a a
readSymbolConcrete :: 
  (MonadPlus m, Eq s) => s -> ReaderArrT r (StateArrT [s] (Kleisli m)) a a
showSymbolConcrete s = MonadArrT (return (StateArrT (second (arr (s:)))))
readSymbolConcrete s = MonadArrT (return (StateArrT (second (Kleisli (readToken s)))))

type RSConcrete' m = ReaderArrT Int (StateArrT [String] (Kleisli m))
type RSConcrete = RSConcrete' ([])

--instance MonadPlus m => ArrowSymbol (RSConcrete' m) where
instance ArrowSymbol RSConcrete where
  showSymbol = showSymbolConcrete
  readSymbol = readSymbolConcrete

-- Idealized ArrowSymbol
readSymbolI :: Eq a => a -> StateArrT [a] (Kleisli ((->) String)) b b
readSymbolI sym = updateStateArrT (Kleisli f)
  where f = \s-> case s of
                    c:cs | sym == c -> return cs
                    []              -> fail

showSymbolI :: StateArrow [a] b => a -> b c c
showSymbolI sym = update (sym:)

-- A concrete ArrowPrec instance

setPrecConcrete :: Prec -> ReaderArrT Prec a b c -> ReaderArrT Prec a b c
setPrecConcrete pri = pokeConcrete pri

pokeConcrete :: r -> ReaderArrT r q a b -> ReaderArrT c q a b
pokeConcrete x (MonadArrT g) = MonadArrT (const (g x))

comapEnv :: (b->c) -> ReaderArrT c d e f -> ReaderArrT b d e f
comapEnv f (MonadArrT g) = MonadArrT (g . f)

mayReadParenConcrete b f = ifHighPrecConcrete b (deparenthesize f) f
mayShowParenConcrete b f = ifHighPrecConcrete b   (parenthesize f) f

ifHighPrecConcrete :: ArrowChoice q => Prec -> ReaderArrT Prec q b c -> ReaderArrT Prec q b c -> ReaderArrT Prec q b c
ifHighPrecConcrete pri = ifConcrete (peekConcrete >>> arr (>pri))

makeReader :: (r -> q a b) -> ReaderArrT r q a b
makeReader f = MonadArrT f

mayReadParenHaskell p f = ifHighPrecHaskell p (deparenthesize f) f
mayShowParenHaskell p f = ifHighPrecHaskell p   (parenthesize f) f
ifHighPrecHaskell :: ArrowChoice q => Prec -> 
  ReaderArrT Prec q b c -> 
  ReaderArrT Prec q b c -> 
  ReaderArrT Prec q b c
ifHighPrecHaskell p = ifConcrete (peekConcrete >>> arr (>p))

peekConcrete :: Arrow q => ReaderArrT r q a r
peekConcrete = MonadArrT (\r-> arr (const r))

ifConcrete :: ArrowChoice q => q a Bool -> q a b -> q a b -> q a b
ifConcrete p t e = arr dup >>> first p >>> arr (uncurry bool2Either) >>> (t ||| e)
  where dup a = (a,a)

-- versions that conform more closely to Haskells show/read
mayReadParenRSWithExtraParens :: 
  (ArrowChoice a, ArrowPlus a, ReaderArrow Prec a, ArrowSymbol a) => 
  Prec -> a b c -> a b c
mayReadParenRSWithExtraParens b f = ifHighPrec b (deparenthesize f') f'
  where f' = mayActuallyReadParen f

mayActuallyReadParen :: (ArrowPlus q, ArrowSymbol q) => q a b -> q a b
mayActuallyReadParen f = f <+> deparenthesize (mayActuallyReadParen f) 

--readSymbolRS " " = readSpaceSymbolRSquickanddirty
--readSymbolRS sym = readNonSpaceSymbolRSquickanddirty sym
-- The proof below is for the quick and dirty versions
--   as the prelude lex is intractable (73 lines of Haskell code)

readSpaceSymbolRS :: Arrow q => q a a
readSpaceSymbolRS = arr id
-- do nothing - assume that readSymbol sym | sym/=" " takes care of white space
--   and does lexing properly to avoid Nilk to be accepted as Nil followed by k
readNonSpaceSymbolRS sym = 
  RS (lift (updateStateArrT (Kleisli (readNonSpaceSymbol sym))))

readNonSpaceSymbol :: String -> String -> [String]
readNonSpaceSymbol sym = lex >>> (filter ((sym==).fst)) >>> map snd

readToken :: (MonadPlus a, Eq b) => b -> [b] -> a [b]
readToken t = 
  \l-> if not (null l) && head l == t
       then return (tail l)
       else mzero

-- quick and dirty: 
--   remove exactly one space character (does not accept \n, \t, etc.)
readSpaceSymbolRSquickanddirty = 
  RS (lift (updateStateArrT (Kleisli readSpaceSymbolquickanddirty)))
readSpaceSymbolquickanddirty :: MonadPlus m => String -> m String
readSpaceSymbolquickanddirty = readToken ' '
-- quick and dirty lexing: just drop white space and take length sym chars
--   roughly twice as fast as using the prelude lex (in Hugs reductions)
readNonSpaceSymbolRSquickanddirty :: String -> RS a a
readNonSpaceSymbolRSquickanddirty sym = 
  RS (lift (updateStateArrT (Kleisli (readNonSpaceSymbolquickanddirty sym))))

readNonSpaceSymbolquickanddirty :: MonadPlus m => String -> String -> m String
readNonSpaceSymbolquickanddirty sym = 
  \s-> let (start,rest) = splitAt (length sym) (dropWhile isSpace s)
       in if start == sym 
          then return rest
          else mzero

-- remove at least one whitespace character
readSpaceSymbolRS2 :: RS a a
readSpaceSymbolRS2 = RS (lift (updateStateArrT (Kleisli readSpaceSymbol2)))

readSpaceSymbol2 :: MonadPlus m => String -> m String
readSpaceSymbol2 =
 \s-> if not (null s) && isSpace (head s)
      then return (dropWhile isSpace s)
      else mzero

